import unittest
from unittest.mock import MagicMock, patch
from numpy.testing import assert_array_equal

import struct
import numpy as np
from scipy.constants import h, c, e

from txrm2tiff import txrm_wrapper


def create_ole_that_returns_integer():
    ole = MagicMock()
    stream = MagicMock()
    ole.openstream.return_value = stream

    packed_pixel_size = struct.pack('<I', 100)
    stream.getvalue.side_effect = [packed_pixel_size]
    return ole


def create_ole_that_returns_float():
    ole = MagicMock()
    stream = MagicMock()
    ole.openstream.return_value = stream

    packed_pixel_size = struct.pack('<f', 100.5)
    stream.getvalue.side_effect = [packed_pixel_size]
    return ole


def pack_int(number):
    return struct.pack('<I', np.int(number))

class TestTxrmWrapper(unittest.TestCase):
    def test_extracting_single_image_short(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.openstream.return_value = stream
        ole.exists.side_effect = [True, False]
        stream.getvalue.return_value = struct.pack('<6H', *list(range(6)))

        data = txrm_wrapper.extract_single_image(ole, 1, 2, 3)

        ole.openstream.assert_called_with('ImageData1/Image1')

        assert_array_equal(data, np.arange(6).reshape(2, 3), err_msg="output is not as expected")

    def test_extracting_single_image_float(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.openstream.return_value = stream
        ole.exists.side_effect = [True, False]
        stream.getvalue.return_value = struct.pack('<6f', *list(range(6)))

        data = txrm_wrapper.extract_single_image(ole, 1, 2, 3)

        ole.openstream.assert_called_with('ImageData1/Image1')

        assert_array_equal(data, np.arange(6).reshape(2, 3), err_msg="output is not as expected")

    def test_extracting_single_image_unexpected_type(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.openstream.return_value = stream
        ole.exists.side_effect = [True, False]
        stream.getvalue.return_value = struct.pack('<6q', *list(range(6)))

        with self.assertRaises(TypeError):
            data = txrm_wrapper.extract_single_image(ole, 1, 2, 3)
            ole.openstream.assert_called_with('ImageData1/Image1')
            ole.openstream.assert_called_with('ImageInfo/DataType')

    def test_read_stream_failure(self):
        ole = MagicMock()
        ole.exists.return_value = False
        data = txrm_wrapper.read_stream(ole, "key", 'i')

        self.assertFalse(data)


    def test_extracts_dimensions(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.openstream.return_value = stream
        stream.getvalue.side_effect = [pack_int(i) for i in [6, 7]]

        data = txrm_wrapper.extract_image_dims(ole)

        self.assertListEqual(data, [6, 7])


    def test_extracts_number_of_images(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.openstream.return_value = stream
        stream.getvalue.side_effect = [pack_int(i) for i in [9]]

        data = txrm_wrapper.extract_number_of_images(ole)

        self.assertEqual(data, 9)


    def test_extracts_multiple_images(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.exists.side_effect = ([True] * 5) + ([False, True] * 5) + [False]
        ole.openstream.return_value = stream
        dimensions = [pack_int(i) for i in [2, 3]]
        images = [struct.pack('<6H', *list(range(6)))] * 5
        images_taken = [pack_int(5)]
        stream.getvalue.side_effect = dimensions + images_taken + images

        data = txrm_wrapper.extract_all_images(ole)

        self.assertEqual(len(data), 5)


    def test_extracts_multiple_images_with_dtypes(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.exists.side_effect = ([True] * 5) + ([True, True] * 5) + [False]
        ole.openstream.return_value = stream
        dimensions = [pack_int(i) for i in [2, 3]]
        dtype = pack_int(5)
        images_and_dtypes = [struct.pack('<6H', *list(range(6))), dtype] * 5
        images_taken = [pack_int(5)]
        stream.getvalue.side_effect = dimensions + images_taken + images_and_dtypes

        data = txrm_wrapper.extract_all_images(ole)

        self.assertEqual(len(data), 5)



    def test_extracts_tilt_angles(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.openstream.return_value = stream
        packed_tilts = struct.pack('<4f', 1, 2, 3, 4)
        stream.getvalue.side_effect = [packed_tilts]

        data = txrm_wrapper.extract_tilt_angles(ole)
        ole.openstream.assert_called_with('ImageInfo/Angles')
        assert_array_equal(data, np.array([1, 2, 3, 4]))


    def test_extacts_exposure_time_tomo(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.openstream.return_value = stream
        # Testing uneven tilt (more +ve than -ve values):
        packed_exposures = struct.pack('<9f', 1, 2, 3, 4, 5, 6, 7, 8, 9)
        packed_angles = struct.pack('<9f', -3, -2, -1, 0, 1, 2, 3, 4, 5)
        stream.getvalue.side_effect = [packed_exposures, packed_angles]
        ole.exists.return_value = True
        data = txrm_wrapper.extract_exposure_time(ole)

        self.assertEqual(data, 4.)


    def test_extacts_exposure_time_single_image(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.openstream.return_value = stream
        packed_exposure = struct.pack('<1f', 5)
        stream.getvalue.side_effect = [packed_exposure]
        ole.exists.side_effect = [False, True, True]
        data = txrm_wrapper.extract_exposure_time(ole)

        self.assertEqual(data, 5)


    def test_extacts_pixel_size(self):
        ole = create_ole_that_returns_float()
        data = txrm_wrapper.extract_pixel_size(ole)

        self.assertEqual(data, 100.5)


    def test_extracts_xray_magnification(self):
        ole = create_ole_that_returns_float()
        data = txrm_wrapper.extract_xray_magnification(ole)

        self.assertEqual(data, 100.5)


    def test_extracts_energy(self):
        ole = create_ole_that_returns_float()
        data = txrm_wrapper.extract_energy(ole)

        self.assertEqual(data, 100.5)


    def test_extracts_wavelength(self):
        ole = create_ole_that_returns_float()
        data = txrm_wrapper.extract_wavelength(ole)
        self.assertAlmostEqual(float("%.8e" % data), 1.23367361e-8)


    def test_create_mosaic_of_reference_image(self):
        reference_data = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

        mosaic_reference = txrm_wrapper.create_reference_mosaic(MagicMock(), reference_data, 6, 3, 2, 1)
        expected_reference = np.array([[1, 2, 3],
                                    [4, 5, 6],
                                    [7, 8, 9],
                                    [1, 2, 3],
                                    [4, 5, 6],
                                    [7, 8, 9]
                                    ])
        self.assertEqual(mosaic_reference.shape, expected_reference.shape, msg="Arrays must be the same shape")
        assert_array_equal(mosaic_reference, expected_reference)


    def test_rescale_ref_exposure(self):
        ole = MagicMock()
        stream = MagicMock()
        ole.openstream.return_value = stream
        # Testing uneven tilt (more +ve than -ve values):
        ref_exposure = struct.pack('<1f', 2)
        packed_exposures = struct.pack('<9f', 1, 2, 3, 4, 5, 6, 7, 8, 9)
        packed_angles = struct.pack('<9f', -3, -2, -1, 0, 1, 2, 3, 4, 5)
        stream.getvalue.side_effect = [ref_exposure, packed_exposures, packed_angles]
        ole.exists.return_value = True
        # In this case, the values should be returned as doubled due to the reference
        # exposure being half of the central image exposure
        assert_array_equal(
            txrm_wrapper.rescale_ref_exposure(ole, np.array([2., 4., 6.])),
            np.array([4., 8., 12.]))

if __name__ == '__main__':
    unittest.main()



import unittest

import re
from pathlib import Path
from datetime import datetime
from tempfile import NamedTemporaryFile

from txrm2tiff.file_sorting import file_can_be_opened, ole_file_works

class TestFileSorting(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.fake_temp = NamedTemporaryFile(suffix=".txrm", delete=False)
        cls.fake_file = Path(cls.fake_temp.name)
        cls.real_file = Path("/dls/science/groups/das/ExampleData/B24_test_data/data/2019/cm98765-1/raw/XMv10/test_tomo2_e3C_full.txrm")

    @classmethod
    def tearDownClass(cls):
        cls.fake_temp.file.close()
        cls.fake_file.unlink()

    def test_ole_file_returns_true_for_real_file(self):
        if self.real_file.exists():
            self.assertTrue(ole_file_works(self.real_file))
        else:
            print("Cannot run test without access to dls directories")

    def test_real_file_can_be_opened(self):
        if self.real_file.exists():
            self.assertTrue(file_can_be_opened(self.real_file))
        else:
            print("Cannot run test without access to dls directories")

    def test_ole_file_returns_false_for_fake_file(self):
        self.assertFalse(ole_file_works(self.fake_file))

    def test_fake_file_can_be_opened(self):
        self.assertTrue(file_can_be_opened(self.fake_file))

    def test_nonexistent_file_can_be_opened(self):
        self.assertFalse(file_can_be_opened(Path("/fake/path/oh/no/{}".format(datetime.now().strftime('%Y%m%d_%H%M')))))


if __name__ == "__main__":
    unittest.main()


from os import access, R_OK
from pathlib import Path
from olefile import OleFileIO, isOleFile
import logging

from .txrm_wrapper import read_imageinfo_as_int



def file_can_be_opened(path):
    if access(str(path), R_OK):
        return True
    logging.error("File %s cannot be opened", path)
    return False


def ole_file_works(path):
    if (path.suffix == ".txrm") or (path.suffix == ".xrm"):
        if isOleFile(str(path)):
            with OleFileIO(str(path)) as ole_file:
                number_frames_taken = read_imageinfo_as_int(ole_file, "ImagesTaken")
                expected_number_frames = read_imageinfo_as_int(ole_file, "NoOfImages")
                # Returns true even if all frames aren't written, throwing warning.
                if number_frames_taken != expected_number_frames:
                    logging.warning("%s is an incomplete %s file: only %i out of %i frames have been written",
                                path.name, path.suffix, number_frames_taken, expected_number_frames)
                # Check for reference frame:
                if not ole_file.exists("ReferenceData/Image"):
                    logging.warning("No reference data found in file %s", path)
                return True
        else:
            logging.warning("Could not read ole file %s", path)
    else:
        logging.warning("%s not .txrm or .xrm", path)
    return False




# -*- coding: utf-8 -*-

from xml.etree import ElementTree
import datetime
import logging
from pathlib import Path

from olefile import OleFileIO, isOleFile
import numpy as np
import tifffile as tf
from oxdls import OMEXML

from . import txrm_wrapper
from .annotator import Annotator

from .__init__ import __version__

dtype_dict = {
    'uint16': np.uint16,
    'float32': np.float32,
    'float64': np.float64
}

def _dynamic_despeckle_and_average_series(images, average=True):
    """
    Despeckle and average a series of images (requires min 3 images)
    This uses the same or similar logic to Xradia/Zeiss' API.

    Args:
        images (list): List of 2D numpy arrays
        average (bool, optional): Defaults to True.

    Returns:
        numpy array: single image (2D array) containing despeckle and averaged data
    """
    nimages, height, width = images.shape
    if nimages < 3:
        logging.error('Despeckle averaging requires a minimum of 3 images')
        return None
    # Takes a list of XradiaData.FloatArray as input
    # outputs a 2D ndarray
    block = images.astype(np.float32).reshape((nimages, height * width))
    # Sort pixels stack-wise
    vals = np.sort(block, axis=0)
    # Set number of sorted frames to split off
    maligned_idxs = [nimages-1]
    if nimages >= 9:
        maligned_idxs.extend([nimages-2, nimages-3, 0])
        good_vals = vals[1: -3]
    else:
        good_vals = vals[:-1]
    maligned = vals[maligned_idxs, :]
    # Calculate mean and sigma of 'good' pixels
    mean = np.mean(good_vals, axis=0, keepdims=True)
    sigma = np.std(good_vals, axis=0, ddof=1, keepdims=True)
    # Create boolean mask for cutting out 'bad' pixels
    bad_pixels = np.fabs(maligned - mean) > 2.8 * sigma
    # Set bad pixels to nan and apply to vals
    maligned[bad_pixels] = np.nan
    vals[maligned_idxs] = maligned
    # Get stack-wise mean (ignoring invalid nans)
    output_image = np.nanmean(vals, axis=0, dtype=np.float32)
    if not average:
        output_image *= nimages
    return output_image.reshape(height, width)


def _apply_reference(images, reference):
    floated_and_referenced = np.asarray(images, dtype=np.float32) * 100. / reference
    if np.isnan(floated_and_referenced).any() or np.isinf(floated_and_referenced).any():
        logging.warning("Potential dead pixels found. "
                        "NaN was output for at least one pixel in the referenced image.")
        # Replace any infinite pixels (nan or inf) with 0:
        _conditional_replace(floated_and_referenced, 0, lambda x: ~np.isfinite(x))
    return floated_and_referenced.astype(np.float32)


def _conditional_replace(array, replacement, condition_func):
    array[condition_func(array)] = replacement


def _get_reference(ole, txrm_name, custom_reference, ignore_reference):
    if custom_reference is not None:
        logging.info("%s is being processed with file %s as a reference.", txrm_name, custom_reference.name)
        reference_path = str(custom_reference)
        try:
            if isOleFile(reference_path):
                with OleFileIO(reference_path) as ref_ole:
                    references = txrm_wrapper.extract_all_images(ref_ole)  # should be float for averaging & dividing
            elif ".tif" in reference_path:
                with tf.TiffFile(reference_path) as tif:
                    references = np.asarray(tif.pages[:])
            else:
                msg = f"Unable to open file '{reference_path}'. Only tif/tiff or xrm/txrm files are supported for custom references."
                logging.error(msg)
                raise IOError(msg)
        except:
            logging.error("Error occurred reading custom reference", exc_info=True)
            raise
        if len(references) > 1:
            # if reference file is an image stack take median of the images
            return _dynamic_despeckle_and_average_series(references)
        return references[0]

    elif ole.exists("ReferenceData/Image") and not ignore_reference:
        logging.info("Internal reference will be applied to %s", txrm_name)
        return txrm_wrapper.extract_reference_image(ole)

    logging.debug("%s is being processed without a reference.", txrm_name)
    return None

def _stitch_images(images, mosaic_xy_shape, fast_axis):
    slow_axis = 1 - fast_axis
    axis_names = ("x", "y")
    logging.debug("Fast axis: %s", axis_names[fast_axis])
    num_images, y, x = images.shape
    expected_num_images = mosaic_xy_shape[0] * mosaic_xy_shape[1]
    
    if expected_num_images > num_images:
        images.resize((expected_num_images, y, x), refcheck=False)

    fast_stacked_list = []
    for i in range(mosaic_xy_shape[slow_axis]):
        idx_start = mosaic_xy_shape[fast_axis] * i
        idx_end = mosaic_xy_shape[fast_axis] * (i + 1)
        logging.debug("Stacking images %i to %i along axis %s", idx_start, (idx_end - 1), axis_names[fast_axis])
        fast_stacked_list.append(np.concatenate(images[idx_start: idx_end], axis=fast_axis))
    # Must be output in a 3D numpy array (with axis 0 as z)
    final_image = np.concatenate(fast_stacked_list, axis=slow_axis)[np.newaxis, :, :]
    logging.info("Mosaic stitched into image with shape: %s", final_image.shape)
    return final_image

def manual_save(tiff_file, image, data_type=None, metadata=None):
    tiff_path = Path(tiff_file)
    image = np.asarray(image)
    image = _cast_to_dtype(image, data_type)
    if metadata is not None:
        meta_img = metadata.image()
        meta_img.Pixels.set_PixelType(str(image.dtype))
        meta_img.set_Name(tiff_path.stem)
        metadata = metadata.to_xml().encode()

    tiff_dir = tiff_path.resolve().parent
    tiff_dir.mkdir(parents=True, exist_ok=True)
    num_frames = len(image)
    bigtiff = image.size * image.itemsize >= np.iinfo(np.uint32).max  # Check if data bigger than 4GB TIFF limit

    logging.info("Saving image as %s with %i frames", tiff_path.name, num_frames)

    with tf.TiffWriter(str(tiff_path), bigtiff=bigtiff, ome=False) as tif:
        tif.write(
            image,
            photometric='MINISBLACK',
            description=metadata,
            metadata={'axes': 'ZYX'},
            software=f"txrm2tiff {__version__}"
            )
    
def save_colour(tiff_file, image):
    tiff_path = Path(tiff_file)
    num_frames = len(image)
    logging.info("Saving annotated image as %s with %i frames", tiff_path.name, num_frames)
    bigtiff = image.size * image.itemsize >= np.iinfo(np.uint32).max  # Check if data bigger than 4GB TIFF limit
    with tf.TiffWriter(str(tiff_path), bigtiff=bigtiff, ome=False) as tif:
        tif.write(
            image,
            photometric='RGB',
            metadata={'axes':'ZYXC'},
            software=f"txrm2tiff {__version__}"
            )


def _convert_output_path_to_annotated_path(output_path):
    """
    Args:
        output_path (pathlib.Path): output path to add "_Annotated" to
    """""
    split_name = output_path.name.split(".")
    num_parts = len(split_name)
    name_idx = num_parts
    suffix = ""
    if num_parts > 1:
        suffix = f".{split_name[-1]}"
        name_idx -= 1
        if num_parts > 2 and split_name[-2].lower() == "ome":
            name_idx -= 1
    annotated_name = ".".join([s for s in split_name[:name_idx]])
    annotated_name += f"_Annotated{suffix}"
    return output_path.parent / annotated_name


def normalise_to_datatype(array, datatype, clip=False):
    logging.debug("Re-scaling array to %s", datatype)
    array = array.astype(np.float64)  # Convert to float for rescaling

    # Clip outliers before normalisation to avoid any speckles
    # that can lead to the normalised image to be dark:
    if clip:
        num_std = 3.
        logging.info("Clipping outliers (>%g std from mean)", num_std)
        new_max = np.mean(array) + num_std * np.std(array)
        np.clip(array, 0, new_max, out=array)

    # Move minimum value of all corrected images to 0:
    array -= array.min()
    # New max should be the max allowed by datatype
    new_max = np.iinfo(datatype).max
    return array * (new_max / array.max())


def _cast_to_dtype(image, data_type):
    if data_type is not None:
        try:
            dtype = np.dtype(data_type)
            if np.issubdtype(dtype, np.integer) and np.issubdtype(image.dtype, np.floating):
                dtype_info = np.iinfo(dtype)
                # Round min/max to avoid this warning when the issue is just going to be rounded away.
                img_min, img_max = round(image.min()), round(image.max())
                if dtype_info.min > img_min:
                    logging.warning(
                        "Image min %f below %s minimum of %i, values below this will be cut off",
                        img_min, dtype, dtype_info.min)
                if dtype_info.max < img_max:
                    logging.warning(
                        "Image max %f above %s maximum of %i, values above this will be cut off",
                        img_max, dtype, dtype_info.max)
                np.clip(image, dtype_info.min, dtype_info.max, out=image)
            return np.around(image, decimals=0).astype(dtype)
        except Exception:
            logging.error("Invalid data type given: %s aka %s. Saving with default data type.", data_type, dtype)
    else:
        logging.warning("No data type specified. Saving with default data type.")
    return image


def create_ome_metadata(ole, image_list, filename=None):
    # Get image dimensions
    # X, Y, number of frames (T in tilt series):
    dimensions = (*image_list[0].shape, len(image_list))
    str_dtype = str(image_list[0].dtype)

    # Get metadata variables from ole file:
    exposures = txrm_wrapper.extract_multiple_exposure_times(ole)

    pixel_size = txrm_wrapper.extract_pixel_size(ole) * 1.e3  # micron to nm

    physical_img_sizes = []
    physical_img_sizes.append(pixel_size * dimensions[0])
    physical_img_sizes.append(pixel_size * dimensions[1])

    x_positions = [coord * 1.e3 for coord in txrm_wrapper.extract_x_coords(ole)]  # micron to nm
    y_positions = [coord * 1.e3 for coord in txrm_wrapper.extract_y_coords(ole)]  # micron to nm

    date_time = datetime.datetime.now().isoformat()  # formatted as: "yyyy-mm-ddThh:mm:ss"
    ox = OMEXML()

    image = ox.image()
    image.set_ID("0")
    image.set_AcquisitionDate(date_time)
    if filename:
        image.set_Name(filename)

    pixels = image.Pixels
    pixels.set_DimensionOrder("XYZCT")
    pixels.set_ID("0")
    pixels.set_PixelType(str_dtype)
    pixels.set_SizeX(dimensions[0])
    pixels.set_SizeY(dimensions[1])
    pixels.set_SizeT(dimensions[2])
    pixels.set_SizeZ(1)
    pixels.set_SizeC(1)
    pixels.set_PhysicalSizeX(physical_img_sizes[0])
    pixels.set_PhysicalSizeXUnit("nm")
    pixels.set_PhysicalSizeY(physical_img_sizes[1])
    pixels.set_PhysicalSizeYUnit("nm")

    pixels.set_plane_count(dimensions[2])
    pixels.set_tiffdata_count(dimensions[2])

    channel = pixels.Channel(0)
    channel.set_ID("Channel:0:0")
    channel.set_Name("C:0")

    if (ole.exists("ImageInfo/MosiacRows") and ole.exists("ImageInfo/MosiacColumns")):
        mosaic_rows = txrm_wrapper.read_imageinfo_as_int(ole, "MosiacRows")
        mosaic_columns = txrm_wrapper.read_imageinfo_as_int(ole, "MosiacColumns")
        # Checks whether the output is a single frame and a mosaic
        # (mosaics should be stitched, and therefore one layer, by now):
        if (mosaic_rows > 0 and mosaic_columns > 0 and dimensions[2] == 1):
            # Calculates:
            # - Mean exposure, throwing away any invalid 0 values
            # - The centre of the stitched mosaic image (as opposed to the centre of a single tile)
            # Both should be returned as a list to reduce changes to the next section.
            exposures = [np.mean([exp for exp in exposures if exp != 0])]
            # The mosaic centre is found by taking the first x & y positions (centre of the first tile,
            # which is the bottom-left in the mosaic), taking away the distance between this and the
            # bottom-left corner, then adding the distance to the centre of the mosaic (calculated using pixel size).
            #
            # More verbosely:
            # The physical size of the stitched mosaic is divided by the rows/columns (columns for x, rows for y).
            # This finds the physical size of a single tile. This is then halved, finding in the physical
            # distance (x, y) from between a corner and the centre of a tile. Then this distance is taken from the
            # (x, y) stage coordinates of the first tile to get the stage coordinates of the bottom-left of the mosaic.
            # Half the physical size of the stitched mosaic is added to this, resulting in the in stage coordinates
            # of the mosaic centre.
            x_positions = [x_positions[0] + (1. - 1. / mosaic_columns) * (physical_img_sizes[0] / 2.)]
            y_positions = [y_positions[0] + (1. - 1. / mosaic_rows) * (physical_img_sizes[1] / 2.)]
            # # NOTE: the number of mosaic rows & columns and the pixel size are all written before acquisition but
            # the xy positions are written during, so only the first frame can be relied upon to have an xy
            # position.

    # Run checks to make sure the value lists are long enough
    exp_len_diff = dimensions[2] - len(exposures)
    if exp_len_diff > 0:
        logging.error("Not enough exposure values for each plane (%i vs %i). Adding zeros to the later planes.",
                        len(exposures), dimensions[2])
        for _ in range(exp_len_diff):
            exposures.append(0)

    x_len_diff = dimensions[2] - len(x_positions)
    if x_len_diff > 0:
        logging.error("Not enough x values for each plane (%i vs %i). Adding zeros to the later planes.",
                        len(x_positions), dimensions[2])
        for _ in range(x_len_diff):
            x_positions.append(0)

    y_len_diff = dimensions[2] - len(y_positions)
    if y_len_diff > 0:
        logging.error("Not enough y values for each plane (%i vs %i). Adding zeros to the later planes.",
                        len(y_positions), dimensions[2])
        for _ in range(y_len_diff):
            y_positions.append(0)

    # Add plane/tiffdata for each plane in the stack
    for count in range(dimensions[2]):
        plane = pixels.Plane(count)
        plane.set_ExposureTime(exposures[count])
        plane.set_TheZ(count)
        plane.set_TheC(0)
        plane.set_TheT(0)

        plane.set_PositionXUnit("nm")
        plane.set_PositionYUnit("nm")
        plane.set_PositionZUnit("reference frame")
        plane.set_PositionX(x_positions[count])
        plane.set_PositionY(y_positions[count])
        plane.set_PositionZ(count)

        tiffdata = pixels.tiffdata(count)
        tiffdata.set_FirstC(0)
        tiffdata.set_FirstZ(count)
        tiffdata.set_FirstT(0)
        tiffdata.set_IFD(count)
        tiffdata.set_plane_count(1)

    return ox

class TxrmToImage:

    def __init__(self):
        self.image_output = None
        self.annotator = None
        self.ome_metadata = None

    def convert(self, txrm_file, custom_reference=None, ignore_reference=False, annotate=False):
        with OleFileIO(str(txrm_file)) as ole:
            images = txrm_wrapper.extract_all_images(ole)
            reference = _get_reference(ole, txrm_file.name, custom_reference, ignore_reference)
            if reference is not None:
                self.image_output = _apply_reference(images, reference)
            else:
                self.image_output = np.around(images)
            if (len(self.image_output) > 1
                    and ole.exists("ImageInfo/MosiacRows")
                    and ole.exists("ImageInfo/MosiacColumns")):
                mosaic_rows = txrm_wrapper.read_imageinfo_as_int(ole, "MosiacRows")
                mosaic_cols = txrm_wrapper.read_imageinfo_as_int(ole, "MosiacColumns")
                if mosaic_rows != 0 and mosaic_cols != 0:
                    # Version 13 style mosaic:
                    self.image_output = _stitch_images(self.image_output, (mosaic_cols, mosaic_rows), 1)
            if annotate:
                # Extract annotations
                annotator = Annotator(self.image_output[0].shape[::-1])
                if annotator.extract_annotations(ole):  # True if any annotations were drawn
                    self.annotator = annotator
                else:
                    self.annotator = False
            # Create metadata
            self.ome_metadata = create_ome_metadata(ole, self.image_output)

    def get_image_and_metadata(self):
        if self.image_output is None:
            logging.warning("Image has not been converted, returning (None, None)")
        return self.image_output, self.ome_metadata

    def get_annotator(self):
        return self.annotator

    def get_annotated_images(self):
        if self.annotator is not None:
            logging.info("Creating a normalised image to apply annotations to")
            return self.annotator.apply_annotations(
                normalise_to_datatype(self.image_output, np.uint8, True)
            )
        return None

    def save(self, tiff_file, data_type=None):
        tiff_path = Path(tiff_file)
        if (self.image_output is not None) and (self.ome_metadata is not None):
            manual_save(tiff_path, self.image_output, data_type, self.ome_metadata)
            if self.annotator is None:
                pass
            elif not self.annotator:
                logging.error("No annotations or scale bar to save with %s", tiff_path)
            else:
                annotationed_images = self.get_annotated_images()
                if annotationed_images is not None:
                    annotated_path = _convert_output_path_to_annotated_path(tiff_path)
                    save_colour(annotated_path, annotationed_images)
        else:
            logging.error("Nothing to save! Please convert the image first")
            raise IOError("Nothing to save! Please convert the image first")



import logging
from pathlib import Path

from olefile import OleFileIO, isOleFile
import numpy as np
from oxdls import OMEXML

from . import txrm_wrapper

class Inspector(OleFileIO):

    def __init__(self, filename):
        self.fp = None  # Assign fp to avoid errors if invalid OLE file (i.e. not opened)
        path = Path(filename)
        self._filename = path.name
        if path.exists() and isOleFile(filename):
            super().__init__(filename)
            self._output_text = f"\n{self._filename}\n\n"
        else:
            raise IOError("'%s' is not a valid xrm/txrm file" % self._filename)
        if self.fp is None:
            raise IOError("'%s' failed to open for unknown reasons" % self._filename)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)

    def inspect(self, extra=False):
        self._output_text = f"\n{self._filename}\n\n"

        self._output_text += "{0} images of type '{1}' with dimensions: {2}\n".format(
            txrm_wrapper.extract_number_of_images(self),
            txrm_wrapper.extract_image_dtype(self, "ImageInfo")[0],
            ", ".join([str(i) for i in txrm_wrapper.extract_image_dims(self)])
            )
        self._output_text += "Pixel size: {0}μm\n".format(
            txrm_wrapper.extract_pixel_size(self))
        rows = (
            txrm_wrapper.read_imageinfo_as_int(self, "MosiacRows")
            if self.exists("ImageInfo/MosiacRows") else 0)
        columns = (
            txrm_wrapper.read_imageinfo_as_int(self, "MosiacColumns")
            if self.exists("ImageInfo/MosiacColumns") else 0)
        # Is a mosaic check:
        if rows > 1 or columns > 1:
            self._output_text += f"Is a mosaic of shape (rows x coumns): {rows}x{columns}\n"
        else:
            self._output_text += "Not a mosaic\n"
        if self.exists("ReferenceData/Image"):
            self._output_text += "Reference of type '{0}' applied with dimensions: {1}\n".format(
                txrm_wrapper.extract_image_dtype(self, "ReferenceData")[0],
                ", ".join([str(i) for i in txrm_wrapper.extract_ref_dims(self)])
                )
        else:
            self._output_text += "No reference applied\n"
        self._output_text += "\n"
        
        if extra:
            axis_dict = txrm_wrapper.get_axis_dict(self)

            self._output_text += "Exposures (s): {0}\n\n".format(
                ", ".join([str(i) for i in txrm_wrapper.extract_multiple_exposure_times(self)]))
            self._output_text += "Xray magnification: {0}\n\n".format(
                txrm_wrapper.extract_xray_magnification(self))

            samplex = axis_dict.get(1)
            if samplex:
                self._output_text += "{0} ({1}): {2}\n\n".format(
                    samplex[0], samplex[1],
                    ", ".join([str(i) for i in txrm_wrapper.extract_x_coords(self)]))
            sampley = axis_dict.get(2)
            if sampley:
                self._output_text += "{0} ({1}): {2}\n\n".format(
                    sampley[0], sampley[1],
                    ", ".join([str(i) for i in txrm_wrapper.extract_y_coords(self)]))
            samplez = axis_dict.get(3)
            if samplez:
                self._output_text += "{0} ({1}): {2}\n\n".format(
                    samplez[0], samplez[1],
                    ", ".join([str(i) for i in txrm_wrapper.extract_z_coords(self)]))
            sampletheta = axis_dict.get(4)
            if sampletheta:
                self._output_text += "{0} ({1}): {2}\n\n".format(
                    sampletheta[0], sampletheta[1],
                    ", ".join([str(i) for i in txrm_wrapper.extract_tilt_angles(self)]))
            energy = axis_dict.get(13)
            if energy:
                self._output_text += "{0} ({1}): {2}\n\n".format(
                    energy[0], energy[1],
                    ", ".join([str(i) for i in txrm_wrapper.extract_energies(self)]))
            self._output_text += "Axis units: {0}\n".format(
                    ", ".join([" ".join(pair) for pair in axis_dict.values()]))
            self._output_text += "\n"

    def _get_stream_list(self):
        return ["/".join(stream) for stream in self.listdir(streams=True, storages=False)]
            

    def list_streams(self):
        for stream in self._get_stream_list():
            self._output_text += f"{stream}\n"

    def inspect_streams(self, *keys):
        self._output_text += "Inspecting streams: " + ", ".join(keys) + "\n"
        try:
            for key in keys:
                if self.exists(key):
                    stream = self.openstream(key)
                    stream_binary = stream.read()

                    self._output_text += f"\n\n{key}:\n"
                    self._output_text += f"\tbinary:           \t{stream_binary}\n"
                    self._output_text += f"\thex binary:       \t{stream_binary.hex()}\n"

                    for buff_type in txrm_wrapper.data_type_dict.values():
                        if buff_type[1] is not None:
                            try:
                                values = ",".join(
                                    [str(i) for i in np.frombuffer(stream.getvalue(), buff_type[1]).tolist()])
                                self._output_text += \
                                    f"\t{buff_type[0] + ':':18s}\t{values}\n"
                            except ValueError as e:
                                self._output_text += f"\t{buff_type[0] + ':':18s}\t{e}\n"
                else:
                    self._output_text += f"{key} does not exist\n"
        except Exception as e:
            self._output_text += f"\n{e}\n"

    def get_text(self):
        return self._output_text



version_info = (1, 3, 1)
__version__ = '.'.join(str(c) for c in version_info)
__author__ = "Thomas Fish"


def convert_and_save(input_path, custom_reference=None, output=None, ignore_reference=False, annotate=False, logging_level="info"):
    """
    Converts xrm or txrm file then saves as it as .ome.tiff file.

    Args:
        input_path (str): Path, can be a file or directory (a directory will batch convert all valid contents).
        custom_reference (str, optional): File path, not valid for batch processing. Defaults to None.
        output (str, optional): Output path, for batch processing output must be a directory otherwise outputs will be saved in the same directory as each input file found. Defaults to None.
        ignore_reference (bool, optional): Ignore any internal reference. Defaults to False.
        annotate (bool, optional): Save additional annotated image (if annotations are available). Defaults to False.
        logging_level (str, optional): Defaults to "info".
    """
    from .run import run
    run(input_path, custom_reference, output, annotate, ignore_reference, logging_level)


def convert(input_file, custom_reference=None, ignore_reference=False):
    """
    Args:
        input_file (str): path to txrm/xrm file to convert
        custom_reference (str, optional): Path to txrm/xrm file to use as reference image. Defaults to None.
        ignore_reference (bool, optional): Ignore any internal reference. Defaults to False.

    Returns:
        list of numpy.ndarrays: list of frames in the image
        omexml-dls object: OME metadata
    """
    from .run import _convert_file
    converter = _convert_file(input_file, custom_reference, ignore_reference, annotate=False)
    return converter.get_image_and_metadata()

def convert_with_annotations(input_file, custom_reference=None, ignore_reference=False):
    """
    Args:
        input_file (str): path to txrm/xrm file to convert
        custom_reference (str, optional): Path to txrm/xrm file to use as reference image. Defaults to None.
        ignore_reference (bool, optional): Ignore any internal reference. Defaults to False.

    Returns:
        numpy.ndarray: image with axis order ZYX
        omexml-dls object: OME metadata
        numpy array or None: Annotated image (if annotations were found) with axis order ZYXC
    """
    from .run import _convert_file
    converter = _convert_file(input_file, custom_reference, ignore_reference, annotate=True)
    output = list(converter.get_image_and_metadata())
    output.append(converter.get_annotated_images())
    return output

def save(output_path, image_list, metadata=None, annotated_image=None):
    """
    Saves image_list as to output formatted as a TIFF. This will be an OME-TIFF if the OME metadata is supplied.
    An annotated image will also be saved as a tiff if the PIL image is suppied.

    Args:
        output_path (str or pathlib.Path): File path to save file (annotated file will have the suffix "_Annotated" applied before the extension)
        image_list (list of numpy.ndarrays): Image to save
        metadata (omexml-dls object, optional): Metadata (will not be applied to annotated image). Defaults to None.
        annotated_image (list of RGB numpy arrays, optional): Annotated images to save. Defaults to None.
    """
    from pathlib import Path
    from .txrm_to_image import manual_save, save_colour

    output_path = Path(output_path)
    manual_save(output_path, image_list, metadata)
    if annotated_image is not None:
        from .txrm_to_image import _convert_output_path_to_annotated_path
        ann_path = _convert_output_path_to_annotated_path(output_path)
        save_colour(annotated_image, ann_path)



import logging
import math
from pathlib import Path
from collections import defaultdict

import numpy as np
from PIL import Image, ImageDraw, ImageFont
from PIL.ImageOps import autocontrast

from . import txrm_wrapper

font_path = Path(__file__).parent / "font" / "CallingCode-Regular.otf"


class Annotator:
    
    def __init__(self, xy_dims):
        """
        Sets up Annotator class for drawing and saving annotations over a converted image.

        Args:
            xy_dims: Tuple or list of dimensions of the image to annotate in the form (x, y)
        """
        self.ann_types = defaultdict(lambda: self._not_implemented, {
            0:  self._plot_line,        # ANN_LINE
            1:  self._plot_rect,        # ANN_RECT
            2:  self._plot_ellipse,     # ANN_ELLIPSE
            3:  self._plot_ellipse,     # ANN_CIRCLE
            # 4:  self._not_implemented,  # ANN_TEXT (Doesn't work in XRMDataExplorer)
            5:  self._plot_polygon,     # ANN_POLYGON
            # 6:  self._not_implemented,  # ANN_GROUP
            # 7:  self._not_implemented,  # ANN_LINE_PLOT
            # 8:  self._not_implemented,  # ANN_RULE
            # 9:  self._not_implemented,  # ANN_ANGLE
            # 10: self._not_implemented,  # ANN_MOUSEPOINT
            # 11: self._not_implemented,  # ANN_CONE_ANGLE_LINE
            # 12: self._not_implemented,  # ANN_HIGHLIGHTER
            13: self._plot_polyline,    # ANN_POLYLINE
            14: self._plot_freehand,    # ANN_FREE_HAND_SKETCH
            # 15: self._not_implemented,  # ANN_SIZE
        })

        self._saved_annotations = False

        self._xy = xy_dims  # Output image dims

        # This set a multiplier so that lines are visible when the image is at a sensible size 
        self._thickness_multiplier = np.mean(self._xy) / 500

        # Create transparent RGBA image with input dims
        self._annotations = Image.new("RGBA", self._xy, (0, 0, 0, 0))
        self._draw = ImageDraw.Draw(self._annotations, mode="RGBA")

    def extract_annotations(self, ole):
        """
        Draws annotations over an RGB copy of the converted image using PIL.

        Args:
            ole: An instance of the xrm/txrm file opened by olefile

        Returns:
            [bool]: True if any annotations have been successfully drawn, otherwise False.
        """
        if ole.exists("Annot/TotalAnn"):
            num_annotations = txrm_wrapper.read_stream(ole, "Annot/TotalAnn", np.intc)[0]
            if num_annotations > 0:
                for i in range(num_annotations):
                    stream_stem = "Annot/Ann%i" %i
                    try:
                        ann_type = txrm_wrapper.read_stream(ole, f"{stream_stem}/AnnType", np.intc)[0]
                        self.ann_types.get(ann_type)(ole, stream_stem)
                        self._saved_annotations = True
                    except Exception:
                        # Create error message but don't stop other annotations from being extracted
                        logging.error("Failed to get annotation %i", i, exc_info=True)
        # _add_scale_bar() returns True if scale bar successfully added
        self._saved_annotations = self._add_scale_bar(ole) or self._saved_annotations
        return self._saved_annotations

    def apply_annotations(self, images):
        """
        Apply annotations to images returning a numpy.ndarray

        Args:
            images (3D NumPy array): Images that will be annotated

        Returns:
            RGB numpy array or None: Annotated images, if successfully applied
        """
        # Create output array with shape of input, plus 3 channels for RGB
        output_images = np.zeros((*images.shape, 3), dtype="uint8")
        try:
            if not self._saved_annotations:
                raise AttributeError("Annotations cannot be as no annotations were successfully extracted")
            images = np.flip(images, 1)
            for idx, image in enumerate(images, 0):
                # Make 'L' type PIL image from 2D array, autocontrast, then convert to RGBA 
                image = autocontrast(Image.fromarray(image).convert("L")).convert("RGBA")
                # Combine annotations and image by alpha
                image.alpha_composite(self._annotations)
                # Throw away alpha and fill into output array
                output_images[idx] = image.convert("RGB")
            return output_images
        except Exception:
            logging.error("Failed to apply annotations", exc_info=True)

    def _add_scale_bar(self, ole):
        pixel_size = txrm_wrapper.extract_pixel_size(ole) # microns
        if pixel_size is not None and pixel_size > 0:
            try:
                colour = (0, 255, 0, 255) # Solid green
                # Set up scale bar dims:
                x0, y0 = self._flip_y((self._xy[0] // 50, self._xy[1] // 30))[0]
                bar_width = int(6 * self._thickness_multiplier)
                # Set up scale bar text:
                f = ImageFont.truetype(str(font_path), int(15 * self._thickness_multiplier))

                tmp_bar_length = self._xy[0] / 5.  # Get initial bar length based on image width
                tmp_bar_size = tmp_bar_length * pixel_size  # Calculate physical size

                # Round to nearest multiple of the order of magnitude
                exponent = math.floor(math.log10(tmp_bar_size))
                step_size = 10. ** exponent
                bar_size = round(
                    round(tmp_bar_size / step_size) * step_size,   # Find nearest multiple of step_size
                    -exponent)  # round to -exponent decimal places (to mitigate rounding errors)
                bar_length = bar_size / pixel_size  # Calculate number of pixels that wide new bar size bar should be
                
                # Set text and calculate text positions:
                text = "%gμm" % bar_size
                text_width, text_height = self._draw.textsize(text, font=f)
                text_x = round(x0 + bar_length / 2 - (text_width / 2))  # Centre text above bar
                text_y = y0 - bar_width - text_height
                # Draw:
                self._draw.text((text_x, text_y), text, font=f, fill=colour)
                self._draw.line((x0, y0, x0 + bar_length, y0), fill=colour, width=bar_width)
                return True
            except Exception:
                logging.error("Exception occurred while drawing scale bar", exc_info=True)
        return False

    def _get_thickness(self, ole, stream_stem):
        return int(txrm_wrapper.read_stream(ole, f"{stream_stem}/AnnStrokeThickness", np.double)[0] * self._thickness_multiplier)

    @staticmethod
    def _get_colour(ole, stream_stem):
        colours = txrm_wrapper.read_stream(ole, f"{stream_stem}/AnnColor", np.ubyte)
        return (*colours[2::-1], colours[3])

    @staticmethod
    def _not_implemented(*args):
        pass

    def _plot_line(self, ole, stream_stem):
        colour = self._get_colour(ole, stream_stem)
        thickness = self._get_thickness(ole, stream_stem)
        x1 = txrm_wrapper.read_stream(ole, f"{stream_stem}/X1", np.single)[0]
        x2 = txrm_wrapper.read_stream(ole, f"{stream_stem}/X2", np.single)[0]
        y1 = txrm_wrapper.read_stream(ole, f"{stream_stem}/Y1", np.single)[0]
        y2 = txrm_wrapper.read_stream(ole, f"{stream_stem}/Y2", np.single)[0]

        self._draw.line(self._flip_y((x1, y1), (x2, y2)), fill=colour, width=thickness)

    def _plot_rect(self, ole, stream_stem):
        colour = self._get_colour(ole, stream_stem)
        thickness = self._get_thickness(ole, stream_stem)
        x0 = txrm_wrapper.read_stream(ole, f"{stream_stem}/Rectangle/Left", np.intc)[0]
        y0 = txrm_wrapper.read_stream(ole, f"{stream_stem}/Rectangle/Bottom", np.intc)[0]
        x1 = txrm_wrapper.read_stream(ole, f"{stream_stem}/Rectangle/Right", np.intc)[0]
        y1 = txrm_wrapper.read_stream(ole, f"{stream_stem}/Rectangle/Top", np.intc)[0]

        self._draw.rectangle(self._flip_y((x0, y0), (x1, y1)), outline=colour, width=thickness)

    def _plot_ellipse(self, ole, stream_stem):
        colour = self._get_colour(ole, stream_stem)
        thickness = self._get_thickness(ole, stream_stem)
        x0 = txrm_wrapper.read_stream(ole, f"{stream_stem}/Rectangle/Left", np.intc)[0]
        y0 = txrm_wrapper.read_stream(ole, f"{stream_stem}/Rectangle/Bottom", np.intc)[0]
        x1 = txrm_wrapper.read_stream(ole, f"{stream_stem}/Rectangle/Right", np.intc)[0]
        y1 = txrm_wrapper.read_stream(ole, f"{stream_stem}/Rectangle/Top", np.intc)[0]
        # Values in the first tuple must be smaller than their respective second tuple value.
        # Therefore y1 must be first and y0 second because they are flipped.
        self._draw.ellipse(self._flip_y((x0, y1), (x1, y0)), outline=colour, width=thickness)

    def _plot_polygon(self, ole, stream_stem):
        colour = self._get_colour(ole, stream_stem)
        thickness = self._get_thickness(ole, stream_stem)
        total_points = txrm_wrapper.read_stream(ole, f"{stream_stem}/TotalPts", np.uintc)[0]
        xs = txrm_wrapper.read_stream(ole, f"{stream_stem}/PointX", np.single)[:total_points]
        ys = txrm_wrapper.read_stream(ole, f"{stream_stem}/PointY", np.single)[:total_points]
        xs.append(xs[0])  # Link beginning and end points
        ys.append(ys[0])
        
        self._draw.line(self._flip_y(*zip(xs, ys)), fill=colour, width=thickness)

    def _plot_polyline(self, ole, stream_stem, joint=None):
        colour = self._get_colour(ole, stream_stem)
        thickness = self._get_thickness(ole, stream_stem)
        total_points = txrm_wrapper.read_stream(ole, f"{stream_stem}/TotalPts", np.uintc)[0]
        xs = txrm_wrapper.read_stream(ole, f"{stream_stem}/PointX", np.single)
        ys = txrm_wrapper.read_stream(ole, f"{stream_stem}/PointY", np.single)

        self._draw.line(self._flip_y(*zip(xs[:total_points], ys[:total_points])), fill=colour, width=thickness, joint=joint)

    def _plot_freehand(self, ole, stream_stem):
        self._plot_polyline(ole, stream_stem, joint="curve")

    def _flip_y(self, *xys):
        """
        Stored x-y coordinates are assuming 0 is bottom left, whereas PIL assumes 0 is top left
        This flips any list of coordinates that alternate x, y

        Args:
            *xys: n args assuming the form (x0, y0), (x1, y1), ... (xn, yn)

        Returns:
            tuple of tuples: x-y coordinates with a flipped y
        """
        xys = tuple((x, self._xy[1] - 1 - y) for (x, y) in xys)
        return xys



from pathlib import Path, PurePath
import logging

from .file_sorting import file_can_be_opened, ole_file_works
from .logger import create_logger
from .txrm_to_image import TxrmToImage


def run(input_path, custom_reference=None, output_path=None, annotate=False, data_type=None, ignore_reference=False, logging_level="info"):
    create_logger(str(logging_level).lower())
    logging.debug(
        "Running with arguments: "
        "input_path=%s, custom_reference=%s, output_path=%s, annotate=%s, data_type=%s, ignore_reference=%s, logging_level=%s",
        input_path, custom_reference, output_path, annotate, data_type, ignore_reference, logging_level)
    input_filepath = Path(input_path)
    if input_filepath.exists():
        if input_filepath.is_dir():
            logging.info("Converting files in directory '%s'", input_path)
            _batch_convert_files(input_filepath, output_path, annotate, data_type, ignore_reference)
        else:
            logging.info("Converting file '%s'", input_path)
            _convert_and_save(input_filepath, output_path, custom_reference, annotate, data_type, ignore_reference)
    else:
        logging.error("No such file or directory: %s", input_path)
        raise IOError(f"No such file or directory: {input_path}")


def _batch_convert_files(input_filepath, output, annotate, data_type, ignore_reference):
    filepath_list = list(input_filepath.rglob("*xrm"))
    logging.info(f"Batch converting files {[filepath.name for filepath in filepath_list]}")
    if output is not None and Path(output).is_dir():
        logging.info(f"Output directory set to: {output}")
        output_path_base = Path(output)
        # Replace input_filepath with output, maintaining sub directory structure, and find output file suffix
        output_path_list = [_define_output_suffix(output_path_base / filepath.relative_to(input_filepath)) for filepath in filepath_list]
        for output_path in output_path_list:
            # Make any directories that do not exist:
            output_path.parent.mkdir(parents=True, exist_ok=True)    
    else:
        # Defines output now rather than in _convert_and_save so that there's a list to iterate
        output_path_list = [_define_output_suffix(filepath) for filepath in filepath_list]
        
    for filepath, output_path in zip(filepath_list, output_path_list):
        _convert_and_save(filepath, output_path, None, annotate, data_type, ignore_reference)


def _convert_and_save(input_filepath, output, custom_reference, annotate, data_type, ignore_reference):
    converter = _convert_file(input_filepath, custom_reference, ignore_reference, annotate)
    if converter is not None:
        output_path = _decide_output(input_filepath, output)
        converter.save(output_path, data_type)


def _convert_file(input_filepath, custom_reference, ignore_reference, annotate):
    if file_can_be_opened(input_filepath) and ole_file_works(input_filepath):
        logging.info(f"Converting {input_filepath.name}")
        converter = TxrmToImage()
        converter.convert(input_filepath, custom_reference, ignore_reference, annotate)
        return converter
    logging.error("Invalid input: '%s'", input_filepath)
        

def _decide_output(input_filepath, output):
    # If no output is supplied
    if output is None:
        output_path = input_filepath
    elif isinstance(output, (str, PurePath)):
        # PurePath is the parent class for all pathlib paths
        output_path = Path(output)
        if output_path.suffix == "" or output_path.is_dir():
            output_path.mkdir(parents=True, exist_ok=True)
            output_path = output_path / input_filepath.name
        else:
            # If a valid filepath was given, just return that (avoids double .ome)
            return output_path
    else:
        logging.error(
            "Invalid output specified: %s. The input path will be used to create the output.",
            output)
        output_path = input_filepath
    return _define_output_suffix(output_path, suffix=input_filepath.suffix)

      
def _define_output_suffix(filepath, suffix=None):
    if suffix is None:
        suffix = filepath.suffix
    if suffix == ".txrm":
        return filepath.with_suffix(".ome.tiff")
    elif suffix == ".xrm":
        return filepath.with_suffix(".ome.tif")
    else:
        logging.error("Invalid file extension: %s", suffix)
        raise NameError(f"Invalid file extension: {suffix}")
